
#ifndef _YOPS_FD_CACHE
#define _YOPS_FD_CACHE

#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <pthread.h>
// #include <atomic_ops.h>
#include <time.h>

#define DEFAULT_CACHE_LEN 2213

struct fd_cache_entry {
	int fd;
	ino_t inode;
	pthread_mutex_t lock;
	int use_cnt; /* usage counter */
	time_t lat;  /* last asked time */
};

struct fd_cache {
	int nent;
	struct fd_cache_entry *entry;

	// TODO: stat
	int ncollisions;
	// etc...
};

static inline struct fd_cache *fdc_init(int nent)
{
	int k;
	struct fd_cache *fdc;
	int n = nent;
	if (n <= 0)
		n = DEFAULT_CACHE_LEN;

	fdc = malloc(sizeof(struct fd_cache));
	if (!fdc) {
		perror("malloc()");
		exit(1);
	}
	memset(fdc, 0, sizeof(struct fd_cache));

	fdc->nent = n;
	fdc->entry = malloc(n*sizeof(struct fd_cache_entry));
	if (!fdc->entry) {
		perror("malloc()");
		exit(1);
	}
	memset(fdc->entry, 0, n*sizeof(struct fd_cache_entry));

	for (k = 0; k < n; k++)
		pthread_mutex_init(&fdc->entry[k].lock, NULL);

	return fdc;
}

/* called only in the 'acceptor' thread of the server */
static inline int fdc_acquire_fd(struct fd_cache *fdc, char *name, ino_t key, int *index)
{
	int fd;
	int i = key % fdc->nent; /* simple hash function */

	pthread_mutex_lock(&fdc->entry[i].lock);

	/* we are lucky - there is a match */
	if (fdc->entry[i].inode == key) {
		// printf("FDC: match (%d, %s)\n", i, name);
		fd = fdc->entry[i].fd;
		goto __done;
	}

	/* collision or empty slot, create new fd */
	fd = open(name, O_RDONLY);
	if (-1 == fd) { /* most likely we ate up all fd's... */
		printf("FDC: can not open '%s'", name); perror("");
		pthread_mutex_unlock(&fdc->entry[i].lock);
		return fd;
	}

	/* is that other file being downloaded currently? */
	if (fdc->entry[i].use_cnt) { /* yes => don't put new fd into the cache */
		printf("FDC: in use (%d, %s)\n", i, name);
		pthread_mutex_unlock(&fdc->entry[i].lock);
		*index = -1; /* mark fd as not being in cache */
		return fd;
	};

	/* 'dirty' entry, slot is free but not empty - close 'old' fd first */
	if (fdc->entry[i].fd) {
		int r;
		r = close(fdc->entry[i].fd);
		if (-1 == r) {
			printf("FDC/BUG: error closing #%d\n", fdc->entry[i].fd);
			perror("close()");
		}
	}

	/* now fill in new fd and inode */
	fdc->entry[i].fd = fd;
	fdc->entry[i].inode = key;

     __done:
	/* there was a match or we added new fd into the cache */

	fdc->entry[i].lat = time(NULL);
	/* increment usage counter */
	fdc->entry[i].use_cnt++;
	printf("++FDC: usage count is %d (fd %d, fdci %d, '%s')\n", fdc->entry[i].use_cnt, fdc->entry[i].fd, i, name);

	pthread_mutex_unlock(&fdc->entry[i].lock);
	*index = i;
	return fd;
}

/* called only in the 'sender' thread of the server */
static inline void fdc_release_fd(struct fd_cache *fdc, char *name, int fd, int index)
{
	/* not in cache, just close */
	if (-1 == index) {
		close(fd);
		return;
	}

	/* file was not found (404)*/
	if (!fd) {
		printf("FDC BUG: fdc_release_fd() get fd #0\n");
		return;
	}

	pthread_mutex_lock(&fdc->entry[index].lock);
	/* update timestamp */
	fdc->entry[index].lat = time(NULL);
	/* decrement usage counter */
	fdc->entry[index].use_cnt--;
	printf("--FDC: usage count is %d (fd %d, fdci %d, '%s')\n", fdc->entry[index].use_cnt, fdc->entry[index].fd, index, name);
	pthread_mutex_unlock(&fdc->entry[index].lock);
}

/* called periodically from the main thread of the server */
static inline void fdc_clear_fd(struct fd_cache *fdc, /*char *name,*/ int index /*, time_t ttl */)
{

	pthread_mutex_lock(&fdc->entry[index].lock);

	/* empty slot */
	if (!fdc->entry[index].fd)
		goto __unlock;

	/* slot is in use */
	if (fdc->entry[index].use_cnt)
		goto __unlock;

	/* entry is not too 'old' yet */
	if ((time(NULL) - fdc->entry[index].lat) < 1800) /* half an hour... TODO: make it configurable */
		goto __unlock;
	/* NOTE: this is MINIMAL time for fd to reside in cache ...
	 * MAXIMAL time is this minimal + delta*fdc->nent, where delta 
	 * is the period of calling this function, see swebs.c:swebs_show_stat()
	 */

	/* entry is old, close and mark as empty */
	close(fdc->entry[index].fd);
	//printf("FDC: entry #%d cleared (fd %d, '%s')\n", index, fdc->entry[index].fd, name);
	printf("FDC: entry #%d cleared (fd %d)\n", index, fdc->entry[index].fd);
	fdc->entry[index].fd = 0;
	fdc->entry[index].inode = 0;

      __unlock:
	pthread_mutex_unlock(&fdc->entry[index].lock);
}

#endif
