
#ifndef SWEBS_H
#define SWEBS_H

#define TCP_PORT 8888
#define SEND_TIMEO_DFLT 30
#define RECV_TIMEO_DFLT 10

#define PIPE_RD_END 0
#define PIPE_WR_END 1

#define CGI_DATA_TIMEOUT 5
#define CGI_MAX_TIME 60

#include <sys/stat.h>
#include <sys/epoll.h>
#include "http.h"

#include "fd-cache.h"

struct job;

struct connection {
	int sock;
	char who[IP_STR_LEN + 1];
	int job_num;
	struct job *job; /* current job */
	int have_request;
};

#define CONG_ARRAY_LEN 256
#define MAX_CONGS (CONG_ARRAY_LEN - 1)
#define MAX_CONG_NAME_LEN 8

/*
 * Fragment format
 * congnm$hash$addr:len<LF>
 */

#define SALT_STRLEN 6
#define HASH_STRLEN 22

/* account for 64 bit addresses */
#define ADDR_STLEN 16

#define JOB_ID_STRLEN (SALT_STRLEN + HASH_STRLEN + ADDR_STLEN + 2)
#define FRAG_LEN_STRLEN 16
#define FRAG_HDR_LEN (JOB_ID_STRLEN + FRAG_LEN_STRLEN + 2)

/* content generator description */
struct cong {
	char name[MAX_CONG_NAME_LEN + 1];
	struct connection *cnx;
	int n;
};

struct job {

	/* filled by acceptor */
	struct connection *cnx;
	int sock;
	time_t when;
	int num;
	char who[IP_STR_LEN + 1];
	struct http_request_header hdr;

	struct cong *cong; /* NULL for static files */
	char id[JOB_ID_STRLEN + 1]; /* to be sent to a cong */
	// unsigned long long int addr_mask;

	/* filled by parser */
	int is_last;
	char content_type[64 + 1];
	char index_file[128 + 1];
	struct stat finfo;

	int response_hlen;
	int response_blen; // HTTP Content-Length
	size_t from_byte;
	size_t upto_byte;
	off_t offs;

	/* filled by launcher */
//	#define NENV 20
//	#define ENV_STR_LEN 256
//	char **argv;
//	char **env;
//	pid_t pid;
//	struct http_response_header cgihdr;

	/* filled by fetcher/launcher, freed by sender */
	int fdci; /* fd cache index, -1 mean not in cache => close after sending file */
	int fd;
	/* pthread_mutex_t lock; */

	unsigned char *buf;
	size_t fetched;
	size_t sent;
	int buf_sent;
	int buf_total;

	int status;
	int done;
	int pollinfd;
	char reason_500[16];
};

struct job_queue {
	char *to_whom;
	int rdfd;
	int wrfd;
	int load; /* number of jobs in the queue */
};

struct worker {
	char *name;
	void* (*action)(void*);

	char *in_queue_name;
	struct job_queue *in;

	char *out_queue_name;
	struct job_queue *out;

	char *aux_queue_name;
	struct job_queue *aux;
};

struct swebs_config {

	unsigned short tcp_port;
	int send_timeo;
	int recv_timeo;

	size_t fragment_size_KiB;

	int cgi_buf_size_KiB;
	int cgi_data_timeo;
	int cgi_max_time;
};

void swebs_check_uids(void);
void swebs_goto_wrkdir(char *reldocroot);
int swebs_open_logfile(char *logname);
void swebs_load_config(char *fname, struct swebs_config *cfg);

void swebs_create_staff(struct worker *workers, int nworkers, struct job_queue *links, int nlinks);

static inline int swebs_listen_on(int port)
{
	int sock;
	sock = soket_server(port, 128);
	if (-1 == sock) {
		perror("socket()");
		exit(1);
	};
	return sock;
}

static inline struct connection *swebs_new_connection(int lsk, struct swebs_config *config, int epoll_fd)
{
	struct timeval send_timeout = {
		.tv_sec = 0,
		.tv_usec = 0
	};

	struct connection *cnx;
	struct epoll_event poll4;
	int sock, r;

	cnx = (struct connection *)insistent_zmalloc(sizeof(struct connection));
	send_timeout.tv_sec = config->send_timeo;
	sock = soket_accept(lsk, cnx->who, &send_timeout);
	if ( -1 == sock) {
		free(cnx);
		return NULL;
	}

	poll4.data.ptr = cnx;
	poll4.events = EPOLLIN;
	r = epoll_ctl(epoll_fd, EPOLL_CTL_ADD, sock, &poll4);
	if (-1 == r) {
		perror("FATAL: epoll_ctl()");
		exit(1);
	};

	cnx->sock = sock;
	return cnx;
}

static inline struct job *swebs_new_job(void)
{
	struct job *job;

	job = insistent_zmalloc(sizeof(struct job));
	time(&job->when);
	// generate job->addr_mask here
	job->hdr.offs = job->hdr.buf;
	return job;
};

static inline int swebs_pass_job(struct job_queue *q, struct job *job)
{
	int pipe = q->wrfd;
	q->load++;
	return write(pipe, &job, sizeof(struct job*));
}

static inline struct job* swebs_get_job(struct job_queue *q)
{
	struct job *job = NULL;
	int r;
	int pipe = q->rdfd;

	r = read(pipe, &job, sizeof(struct job*));
	if (r != sizeof(struct job*))
		return NULL;
	q->load--;
	return job;
};

// use some hash!!!
static inline int swebs_ins_cong(struct cong *congs, char *name, struct connection *cnx)
{
	int k;

	if (MAX_CONGS == congs[0].n)
		return 0;

	/* find first empty slot */
	for (k = 1; k <= MAX_CONGS; k++)
		if (!congs[k].cnx)
			break;

	if (congs[k].cnx)
		return 0;

	congs[k].cnx = cnx;
	congs[k].n = k;
	strcpy(congs[k].name, name);

	/* increment number of congs */
	congs[0].n++;
	return k;
}

static inline int swebs_find_cong(struct cong *congs, char *name)
{
	int cn;
	int found = 0;
	char *needle;
	
	for (cn = 1; cn <= MAX_CONGS; cn++) {
		if (strlen(congs[cn].name)) {
			needle = strstr(name, congs[cn].name);
			if (needle == name) {
				found = 1;
				break;
			}
		}
	}

	if (!found)
		return 0;

	return cn;
}

static inline int swebs_del_cong(struct cong *congs, char *name)
{
	int cn;

	cn = swebs_find_cong(congs, name);
	if (!cn) {
		printf("OOPS! - attempt to delete unexistent CONG (%s)!\n", name);
		return 0;
	}

	memset(&congs[cn], 0, sizeof(struct cong));

	return 0;
}

//int swebs_get_request(struct job *job, struct swebs_config *config);
int swebs_get_request(struct job *job);

int swebs_read_http_request(struct job *job, struct swebs_config *cfg);
int swebs_parse_http_request(struct job *job);
int swebs_parse_uri(struct job *job);
int swebs_send_response(struct job *job);
int swebs_send_builtin_response(struct job *job);

int swebs_prepare_to_load_file(struct job *job, struct fd_cache *fdc, struct swebs_config *cfg);
int swebs_get_fragment(struct job *job, struct swebs_config *cfg);
void swebs_end_send_file(struct job *job, struct fd_cache *fdc);
int swebs_send(struct job *job);

int swebs_set_environment(struct job *job, struct swebs_config *cfg);
int swebs_clear_environment(struct job *job);
pid_t swebs_launch_script(struct job *job, int *pipe, struct swebs_config *config, int errlog);
int swebs_read_response_header(struct job *job, struct swebs_config *cfg);
int swebs_parse_response_header(struct job *job);

int swebs_record_log(int log, struct job *job);
void swebs_show_stat(struct job_queue *links, int nlinks, struct fd_cache *fdc);

void swebs_generate_addr_mask(unsigned long long *ptr);
void swebs_gen_jobid(struct job *job, unsigned long long int mask);
struct job *swebs_check_jobid(char *jobid, unsigned long long int mask);

#define CHECK_AND_EXIT(condition)				\
	if (!(condition)) {					\
		sleep(2);					\
		fprintf(					\
			stdout, 				\
			 "\n!YOOPS! in '%s' @ line %d (%s)\n\n",\
			 __FILE__,				\
			 __LINE__,				\
			 __FUNCTION__				\
		);						\
		exit(1);					\
	}

#define LOG_SYS_ERR(syscall, errn)				\
	do {							\
		time_t ts = time(NULL);				\
		fprintf(stderr,					\
			"[%s]/%s() in '%s' @ line %d at %s",	\
			strerror(errn),				\
			syscall,				\
			__FILE__,				\
			__LINE__,				\
			ctime(&ts)				\
		);						\
	} while (0);

#endif
